/*
 * This file is part of the Nepomuk KDE project.
 * Copyright (C) 2008 Adrien Bustany <madcat@mymadcat.com>
 * Copyright (C) 2008-2009 Sebastian Trueg <trueg@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <QtCore/QVariant>

#include <Nepomuk/Types/Property>
#include <Nepomuk/Types/Class>
#include <Nepomuk/Resource>

#include <KServiceTypeTrader>
#include <KService>
#include <KDebug>

#include "annotationpluginfactory.h"
#include "annotationplugin.h"


class Nepomuk::AnnotationPluginFactory::Private
{
public:
    QHash<QString, AnnotationPlugin*> cachedPlugins;
    AnnotationPlugin* fallback;
};



Nepomuk::AnnotationPluginFactory::AnnotationPluginFactory(QObject* parent)
    : QObject(parent),
      d( new Private )
{
	// Create the base plugin
    d->fallback = new AnnotationPlugin(this);
}


Nepomuk::AnnotationPluginFactory::~AnnotationPluginFactory()
{
    qDeleteAll( d->cachedPlugins );
    delete d;
}


K_GLOBAL_STATIC( Nepomuk::AnnotationPluginFactory, s_globalFactory )

Nepomuk::AnnotationPluginFactory* Nepomuk::AnnotationPluginFactory::instance()
{
    return s_globalFactory;
}

QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getPluginsForProperty(const Nepomuk::Types::Property& property)
{
	QList<AnnotationPlugin*> plugins;
	if(!property.isValid())
		return plugins;

	// Note : There's one service file for one library, but a single service file can contain several properties
	KService::List services = KServiceTypeTrader::self()->query("Nepomuk/AnnotationPlugin", QString("'%1' in [X-KDE-NepomukProperties]").arg(property.uri().toString()));

	foreach(const KSharedPtr<KService>& service, services) {
		// Return the cached instance if we have one, create a new one else
        AnnotationPlugin* plugin = getPluginInstance(service);
        if ( plugin && !plugins.contains( plugin ) )
            plugins.append( plugin );
    }

	return plugins;
}

QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getPluginsForType(const QUrl& rdfType)
{
	QList<AnnotationPlugin*> plugins;

	// Note : There's one service file for one library, but a single service file can contain several properties
	KService::List services
        = KServiceTypeTrader::self()->query( "Nepomuk/AnnotationPlugin",
                                             QString("'%1' in [X-KDE-NepomukResourceType]")
                                             .arg(rdfType.toString()));

	foreach(const KSharedPtr<KService>& service, services) {
		// Return the cached instance if we have one, create a new one else
        AnnotationPlugin* plugin = getPluginInstance(service);
        if ( plugin && !plugins.contains( plugin ) )
            plugins.append( plugin );
    }

    if ( plugins.isEmpty() )
        plugins << d->fallback;

	return plugins;
}


QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getPluginsSupportingAnnotationOfTypes( const QList<Types::Class>& rdfTypes )
{
	QList<AnnotationPlugin*> plugins;

    QStringList subQueries( "'*' in [X-Nepomuk-ResourceTypes]" );
    for( int i = 0; i < rdfTypes.count(); ++i ) {
        subQueries << QString( "'%1' in [X-Nepomuk-ResourceTypes]" ).arg( rdfTypes[i].uri().toString() );
    }

	KService::List services
        = KServiceTypeTrader::self()->query( "Nepomuk/AnnotationPlugin", subQueries.join( " or " ) );

	foreach(const KSharedPtr<KService>& service, services) {
		// Return the cached instance if we have one, create a new one else
        AnnotationPlugin* plugin = getPluginInstance(service);
        if ( plugin && !plugins.contains( plugin ) )
            plugins.append( plugin );
    }

	return plugins;
}


QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getPluginsSupportingAnnotationOfType( const Types::Class& rdfType )
{
    return getPluginsSupportingAnnotationOfTypes( QList<Types::Class>() << rdfType );
}


QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getPluginsSupportingAnnotationForResource( const QUrl& resource )
{
    Nepomuk::Resource res( resource );
    QList<Types::Class> types;
    foreach( const QUrl& type, res.types() ) {
        Types::Class typeClass( type );
        types << typeClass;
        types << typeClass.allParentClasses();
    }
    return getPluginsSupportingAnnotationOfTypes( types );
}


Nepomuk::AnnotationPlugin* Nepomuk::AnnotationPluginFactory::getPluginByLibraryName(const QString& name)
{
	// I couldn't get it to work with the trader, so here's a brutal version
	KService::List services = KServiceTypeTrader::self()->query("Nepomuk/AnnotationPlugin");

	foreach(const KSharedPtr<KService>& service, services) {
		if(service->library() == name)
			return getPluginInstance(service);
	}

	return NULL;
}

QList<Nepomuk::AnnotationPlugin*> Nepomuk::AnnotationPluginFactory::getAllPlugins()
{
	QList<AnnotationPlugin*> plugins;

	KService::List services = KServiceTypeTrader::self()->query("Nepomuk/AnnotationPlugin");

	foreach(const KSharedPtr<KService>& service, services) {
		// Return the cached instance if we have one, create a new one else
        AnnotationPlugin* plugin = getPluginInstance(service);
        if ( plugin && !plugins.contains( plugin ) )
            plugins.append( plugin );
    }

	return plugins;
}

Nepomuk::AnnotationPlugin* Nepomuk::AnnotationPluginFactory::getPluginInstance(const KSharedPtr<KService>& service)
{
	if(d->cachedPlugins.contains(service->library())) {
		return d->cachedPlugins.value(service->library());
	} else {
		AnnotationPlugin* plugin = service->createInstance<AnnotationPlugin>(this, QVariantList(), 0);
		if(!plugin) {
			kDebug() << "Error, cannot instantiate plugin from library" << service->library();
		} else {
			d->cachedPlugins.insert(service->library(), plugin);
			return plugin;
		}
	}
	return 0;
}

#include "annotationpluginfactory.moc"

