/*
   Copyright (c) 2008-2009 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "abstractannotationmodel.h"
#include "annotation.h"
#include "annotationpluginwrapper.h"
#include "annotationplugin.h"
#include "annotationpluginfactory.h"

#include <KIcon>
#include <KDebug>

#include <Nepomuk/Resource>


class Nepomuk::AbstractAnnotationModel::Private
{
public:
    Private( AbstractAnnotationModel* parent )
        : q( parent ) {
    }

    QList<Annotation*> annotations;

    Resource resource;

    void annotationCreated( Nepomuk::Annotation* anno );

private:
    AbstractAnnotationModel* q;
};


void Nepomuk::AbstractAnnotationModel::Private::annotationCreated( Nepomuk::Annotation* anno )
{
    int row = annotations.indexOf( anno );
    if ( row >= 0 ) {
        q->beginRemoveRows( QModelIndex(), row, row );
        annotations.takeAt( row );
        q->endRemoveRows();
    }
    emit q->annotationCreated( anno );

    // FIXME: should we at this point always update annotations? After all, the new annotation could
    //        have changed the state in a way that some annotations do not make sense anymore or new
    //        ones do.
}


Nepomuk::AbstractAnnotationModel::AbstractAnnotationModel( QObject* parent )
    : QAbstractItemModel( parent ),
      d( new Private(this) )
{
}


Nepomuk::AbstractAnnotationModel::~AbstractAnnotationModel()
{
    delete d;
}


int Nepomuk::AbstractAnnotationModel::columnCount( const QModelIndex& ) const
{
    return 1;
}


QVariant Nepomuk::AbstractAnnotationModel::data( const QModelIndex& index, int role ) const
{
    if ( Annotation* annotation = annotationForIndex( index ) ) {
        switch( role ) {
        case Qt::DisplayRole:
            return annotation->comment() + QString( " (%1)" ).arg( annotation->relevance() );

        case Qt::DecorationRole:
            return annotation->icon();

        case RelevanceRole:
            return annotation->relevance();
        }
    }
    return QVariant();
}


QModelIndex Nepomuk::AbstractAnnotationModel::parent( const QModelIndex& ) const
{
    return QModelIndex();
}


int Nepomuk::AbstractAnnotationModel::rowCount( const QModelIndex& parent ) const
{
    return( parent.isValid() ? 0 : d->annotations.count() );
}


QModelIndex Nepomuk::AbstractAnnotationModel::index( int row, int column, const QModelIndex& parent ) const
{
    if ( row < d->annotations.count() ) {
        return createIndex( row, column, d->annotations[row] );
    }
    else {
        return QModelIndex();
    }
}


Nepomuk::Annotation* Nepomuk::AbstractAnnotationModel::annotationForIndex( const QModelIndex& index ) const
{
    if ( index.internalPointer() ) {
        return static_cast<Annotation*>( index.internalPointer() );
    }
    else {
        return 0;
    }
}


Nepomuk::Annotation* Nepomuk::AbstractAnnotationModel::takeAnnotation( const QModelIndex& index )
{
    if ( index.row() < d->annotations.count() ) {
        beginRemoveRows( QModelIndex(), index.row(), index.row() );
        Annotation* a = d->annotations.takeAt( index.row() );
        a->disconnect( this );
        endRemoveRows();
        return a;
    }
    else {
        return 0;
    }
}


void Nepomuk::AbstractAnnotationModel::addNewAnnotation( Annotation* anno )
{
    beginInsertRows( QModelIndex(), d->annotations.count(), d->annotations.count() );
    d->annotations << anno;
    connect( anno, SIGNAL( finished( Nepomuk::Annotation* ) ),
             this, SLOT( annotationCreated( Nepomuk::Annotation* ) ) );
    endInsertRows();
}


void Nepomuk::AbstractAnnotationModel::setResource( const Resource& resource )
{
    d->resource = resource;
    updateAnnotations();
}


void Nepomuk::AbstractAnnotationModel::updateAnnotations()
{
    qDeleteAll( d->annotations );
    d->annotations.clear();
    reset();
    doUpdateAnnotations();
}


Nepomuk::Resource Nepomuk::AbstractAnnotationModel::resource() const
{
    return d->resource;
}

#include "abstractannotationmodel.moc"
