/*
   Copyright (c) 2008 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "geonamesannotationplugin.h"
#include "geonamescountryfetchjob.h"
#include "graphretriever.h"
#include "simpleannotation.h"
#include "pimo.h"
#include "pimoplugintools.h"

#include <Soprano/Model>
#include <Soprano/Node>
#include <Soprano/QueryResultIterator>
#include <Soprano/Statement>
#include <Soprano/StatementIterator>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/NRL>
#include <Soprano/Vocabulary/NAO>
#include <Soprano/Vocabulary/XMLSchema>

#include <Nepomuk/Thing>
#include <Nepomuk/Variant>
#include <Nepomuk/ResourceManager>

#include <KGlobal>
#include <KLocale>
#include <KDebug>
#include <KIcon>

#include <QtCore/QTimer>


namespace {
    // TODO: using SimpleAnnotation statistics here may not make much sense as it is way
    //       more interesting to know if the user uses the geo annotations at all, not the
    //       specific countries. Could we maybe even represent that in the Annotation API?
    class GeoNamesAnnotation : public Nepomuk::SimpleAnnotation
    {
    public:
        GeoNamesAnnotation( const Nepomuk::Resource& resource, QObject* parent )
            : SimpleAnnotation( resource, parent ) {
        }

        void setGeoNamesFeature( const QUrl& resource, const QString& name, const QUrl& featureClass ) {
            m_geoFeature = resource;
            m_geoFeatureName = name;
            m_geoFeatureClass = featureClass;

            // needed for statistics in SimpleAnnotation::occurenceCount and others
            SimpleAnnotation::setValue( m_geoFeature );
        }

    protected:
        void doCreate() {
            QUrl thingType;
            if ( m_geoFeatureClass.fragment() == "P" )
                thingType = Nepomuk::Vocabulary::PIMO::City();
            else
                thingType = Nepomuk::Vocabulary::PIMO::Country();

            Nepomuk::Resource geoFeatureRes( m_geoFeature, QUrl::fromEncoded( "http://www.geonames.org/ontology#Feature", QUrl::StrictMode ) );
            geoFeatureRes.pimoThing().setLabel( m_geoFeatureName );
            geoFeatureRes.pimoThing().addType( thingType );

            Nepomuk::Resource res( resource() );
            QList<Nepomuk::Resource> rl = res.pimoThing().property( property().uri() ).toResourceList();
            rl << geoFeatureRes.pimoThing();
            res.pimoThing().setProperty( property().uri(), rl );

            emitFinished();
        }

    private:
        QUrl m_geoFeature;
        QString m_geoFeatureName;
        QUrl m_geoFeatureClass;
    };

    /**
     * geonames does not handle all language variants such as 'en_US' or 'en_GB' but
     * only the basic language codes, in this example 'en'.
     */
    QString normalizedLanguageCode() {
        QString language = KGlobal::locale()->language();
        int p = language.indexOf( '_' );
        if ( p > 0 ) {
            return language.left( p );
        }
        else {
            return language;
        }
    }
}


Nepomuk::GeoNamesAnnotationPlugin::GeoNamesAnnotationPlugin( QObject* parent, const QVariantList& )
    : AnnotationPlugin( parent ),
      m_lastRdfJob( 0 )
{
    QTimer::singleShot( 0, this, SLOT( fetchCountryInformation() ) );
}


Nepomuk::GeoNamesAnnotationPlugin::~GeoNamesAnnotationPlugin()
{
}


void Nepomuk::GeoNamesAnnotationPlugin::doGetPossibleAnnotations( const Resource& resource, const QString& filter )
{
    // ignore any previous jobs
    if ( m_lastRdfJob ) {
        m_lastRdfJob->disconnect( this );
        m_lastRdfJob = 0;
    }

    if ( filter.length() > 2 ) {
        kDebug() << "starting geonames query for" << filter;

        m_resource = resource;
        m_filter = filter;

        // build geonames query
        QUrl url( "http://ws5.geonames.org/search" );

        // search for names of things that start with filter
        url.addQueryItem( "name", filter + '*' );

        // no more than 5 results
        url.addQueryItem( "maxRows", "5" );

        // default to the user's language if possible
        url.addQueryItem( "lang", normalizedLanguageCode() );

        // get the full info which includes the alt names
        url.addQueryItem( "style", "full" );

        // we only want to handle countries and cities for now
        url.addQueryItem( "featureClass", "P" ); // city, village, ...
        url.addQueryItem( "featureClass", "A" ); // country, state, region, ...

        // and please give us some rdf
        url.addQueryItem( "type", "rdf" );

        kDebug() << url;

        m_lastRdfJob = new GraphRetriever( this );
        connect( m_lastRdfJob, SIGNAL( result( KJob* ) ),
                 this, SLOT( slotGeoNamesQueryResult( KJob* ) ) );
        m_lastRdfJob->setUrl( url );
        m_lastRdfJob->start();
    }
    else {
        emitFinished();
    }
}


void Nepomuk::GeoNamesAnnotationPlugin::slotGeoNamesQueryResult( KJob* job )
{
    if ( !job->error() ) {
        if ( Soprano::Model* model = m_lastRdfJob->model() ) {
            Soprano::QueryResultIterator it = model->executeQuery( QString( "select ?r ?name ?altname ?type ?country where { "
                                                                            "?r a <http://www.geonames.org/ontology#Feature> . "
                                                                            "?r <http://www.geonames.org/ontology#name> ?name . "
                                                                            "?r <http://www.geonames.org/ontology#featureClass> ?type . "
                                                                            "?r <http://www.geonames.org/ontology#inCountry> ?country . "
                                                                            "OPTIONAL { ?r <http://www.geonames.org/ontology#alternateName> ?altname . "
                                                                            "FILTER(LANG(?altname) = '%1') . } }" )
                                                                   .arg( normalizedLanguageCode() ),
                                                                   Soprano::Query::QueryLanguageSparql );
            while ( it.next() ) {
                QUrl featureUri = it.binding( "r" ).uri();
                QString name = it.binding( "name" ).toString();
                QString altname = it.binding( "altname" ).toString();
                QUrl country = it.binding( "country" ).uri();

                // do not create an annotation twice
                if ( !m_resource.pimoThing().isValid() ||
                     !ResourceManager::instance()->mainModel()->executeQuery( QString( "ask where { ?r <%1> <%2> . <%3> <%4> ?r . }" )
                                                                              .arg( Vocabulary::PIMO::groundingOccurrence().toString() )
                                                                              .arg( featureUri.toString() )
                                                                              .arg( m_resource.pimoThing().resourceUri().toString() )
                                                                              .arg( Vocabulary::PIMO::hasLocation().toString() ),
                                                                              Soprano::Query::QueryLanguageSparql ).boolValue() ) {
                    QUrl type = it.binding( "type" ).uri();

                    QString displayName = altname.isEmpty() ? name : altname;
                    if ( type.fragment() != "A" ) {   // makes no sense to have labels like "country (country)"
                        QString countryName = getCountryNameFromInCountryUri( country );
                        if ( !countryName.isEmpty() ) {
                            displayName = QString( "%1 (%2)" ).arg( displayName, countryName );
                        }
                    }

                    GeoNamesAnnotation* anno = new GeoNamesAnnotation( m_resource, this );
                    anno->setProperty( Vocabulary::PIMO::hasLocation() );
                    anno->setGeoNamesFeature( featureUri, altname.isEmpty() ? name : altname, type );
                    anno->setLabel( name );
                    anno->setComment( i18nc( "@item:listbox Statement about a person or something being located in a city or a country",
                                             "is located in %1", displayName ) );
                    anno->setRelevance( 1.0 - calculateDistance( name, m_filter ) );
                    anno->setIcon( KIcon( "nepomuk" ) ); // don't have anything better for now

                    addNewAnnotation( anno );
                }
            }

            delete model;
        }
    }

    m_lastRdfJob = 0;

    emitFinished();
}


// geonames' query service does not provide country links in features. Thus, we simply
// cache all countries in the local Nepomuks store. I don't think that anyone would
// mind having a bit of geographical data at their fingertips. ;)
void Nepomuk::GeoNamesAnnotationPlugin::fetchCountryInformation()
{
    if ( !GeoNamesCountryFetchJob::countriesAlreadyFetched() ) {
        GeoNamesCountryFetchJob* job = new GeoNamesCountryFetchJob( this );
        job->start(); // don't care about result, we can't do anything about it anyway and will retry the next time
    }
}


QString Nepomuk::GeoNamesAnnotationPlugin::getCountryNameFromInCountryUri( const QUrl& uri )
{
    QString countryName;

    // let's see if fetchCountryInformation has already stored the country information
    // here we hack a bit: we have no real country link, only the geonames:inCountry one which
    // we can only match against the inCountry property of the country which is in that country (what?)
    Soprano::QueryResultIterator it
        = ResourceManager::instance()->mainModel()->executeQuery( QString( "select ?name ?altname where { "
                                                                           "?c a <http://www.geonames.org/ontology#Feature> . "
                                                                           "?c <http://www.geonames.org/ontology#featureClass> <http://www.geonames.org/ontology#A> . "
                                                                           "?c <http://www.geonames.org/ontology#inCountry> <%1> . "
                                                                           "?c <http://www.geonames.org/ontology#name> ?name . "
                                                                           "OPTIONAL { ?c <http://www.geonames.org/ontology#alternateName> ?altname . "
                                                                           "FILTER(LANG(?altname) = '%2') . } }" )
                                                                  .arg( QString::fromAscii( uri.toEncoded() ) )
                                                                  .arg( KGlobal::locale()->language() ),
                                                                  Soprano::Query::QueryLanguageSparql );

    if ( it.next() ) {
        countryName = it.binding( "altname" ).toString();
        if ( countryName.isEmpty() )
            countryName = it.binding( "name" ).toString();
    }

    return countryName;
}

NEPOMUK_EXPORT_ANNOTATION_PLUGIN( Nepomuk::GeoNamesAnnotationPlugin, "nepomuk_geonamesannotationplugin" )

#include "geonamesannotationplugin.moc"
