/*
   Copyright (c) 2008 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "geonamescountryfetchjob.h"
#include "graphretriever.h"

#include <Soprano/Model>
#include <Soprano/Node>
#include <Soprano/QueryResultIterator>
#include <Soprano/Statement>
#include <Soprano/StatementIterator>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/NRL>
#include <Soprano/Vocabulary/NAO>
#include <Soprano/Vocabulary/XMLSchema>
#include <Soprano/Util/AsyncModel>
#include <Soprano/Util/AsyncResult>

#include <Nepomuk/ResourceManager>

#include <QtCore/QDateTime>

#include <KLocale>
#include <KDebug>


namespace {
    const int s_countryFetchBatchSize = 50;
}

Nepomuk::GeoNamesCountryFetchJob::GeoNamesCountryFetchJob( QObject* parent )
    : KJob( parent ),
      m_fetchedCountriesCount( 0 )
{
    // we use an async model to add the statements. We will add more than 23000 statemenst
    // and we do not want to block the GUI doing that!
    m_asyncModel = new Soprano::Util::AsyncModel( ResourceManager::instance()->mainModel() );
    m_asyncModel->setMode( Soprano::Util::AsyncModel::MultiThreaded );
}


Nepomuk::GeoNamesCountryFetchJob::~GeoNamesCountryFetchJob()
{
    delete m_asyncModel;
}


bool Nepomuk::GeoNamesCountryFetchJob::countriesAlreadyFetched()
{
    return ResourceManager::instance()->mainModel()->executeQuery( QString( "ask where { ?g a <%1> . ?g <%2> \"geonames/countries\"^^<%3> . }" )
                                                                   .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() )
                                                                   .arg( Soprano::Vocabulary::RDFS::label().toString() )
                                                                   .arg( Soprano::Vocabulary::XMLSchema::string().toString() ),
                                                                   Soprano::Query::QueryLanguageSparql ).boolValue();
}


void Nepomuk::GeoNamesCountryFetchJob::start()
{
    // reset some vars
    m_fetchedCountriesCount = 0;
    m_fetchedCountryStatementsCount = 0;
    m_retryCount = 0;

    // we do an update as in: delete + re-fetch
    removeOldData();

    // store the graph that will contain all the country statements
    m_countriesGraph = ResourceManager::instance()->generateUniqueUri( "geonames-countries" );

    // start the actual fetching
    fetchNextBatch();
}


void Nepomuk::GeoNamesCountryFetchJob::removeOldData()
{
    // get the graph names
    Soprano::QueryResultIterator it
        = ResourceManager::instance()->mainModel()->executeQuery( QString( "select ?g ?m where { "
                                                                           "?g a <%1> . "
                                                                           "?g <%2> \"geonames/countries\"^^<%3> . "
                                                                           "?m <%4> ?g . }" )
                                                                  .arg( Soprano::Vocabulary::NRL::InstanceBase().toString() )
                                                                  .arg( Soprano::Vocabulary::RDFS::label().toString() )
                                                                  .arg( Soprano::Vocabulary::XMLSchema::string().toString() )
                                                                  .arg( Soprano::Vocabulary::NRL::coreGraphMetadataFor().toString() ),
                                                                  Soprano::Query::QueryLanguageSparql );
    if ( it.next() ) {
        QUrl g = it.binding( "g" ).uri();
        QUrl m = it.binding( "m" ).uri();
        it.close();
        ResourceManager::instance()->mainModel()->removeContext( g );
        ResourceManager::instance()->mainModel()->removeContext( m );
    }
}


void Nepomuk::GeoNamesCountryFetchJob::fetchNextBatch()
{
    kDebug() << "fetching batch starting at" << ( m_fetchedCountriesCount+1 ) << "with size" << s_countryFetchBatchSize;

    GraphRetriever* countryFetchJob = new GraphRetriever( this );

    // build geonames query
    QUrl url( "http://ws.geonames.org/search" );

    // we fetch the data in batches since all countries have around 23000 statements and
    // the parser keeps them all in a QList (maybe fix the parser ;)
    url.addQueryItem( "maxRows", QString::number( s_countryFetchBatchSize ) );
    url.addQueryItem( "startRow", QString::number( m_fetchedCountriesCount+1 ) );

    // we want all the available information
    url.addQueryItem( "style", "full" );

    // we only want countries
    url.addQueryItem( "featureClass", "A" ); // country, state, region, ...

    // we only want political entities (this way we miss about 12 countries
    // that have the ADM1 code but those are things like Antarctica which
    // we can live without.)
    // If we would also fetch ADM1 codes we would get a bunch of features
    // that have the same country acronym (about 70)
    url.addQueryItem( "featureCode", "PCL" );
    url.addQueryItem( "featureCode", "PCLI" );
    url.addQueryItem( "featureCode", "PCLD" );
    url.addQueryItem( "featureCode", "PCLS" );
    url.addQueryItem( "featureCode", "PCLF" );
    url.addQueryItem( "featureCode", "PCLIX" );

    // and please give us some rdf
    url.addQueryItem( "type", "rdf" );

    connect( countryFetchJob, SIGNAL( result( KJob* ) ),
             this, SLOT( slotCountriesBatchResult( KJob* ) ) );
    countryFetchJob->setUrl( url );
    countryFetchJob->start();
}


void Nepomuk::GeoNamesCountryFetchJob::slotCountriesBatchResult( KJob* job )
{
    if ( !job->error() ) {
        GraphRetriever* countryFetchJob = static_cast<GraphRetriever*>( job );

        int cnt = 0;

        // put all the country information into one named graph
        Soprano::StatementIterator it = countryFetchJob->statements();
        QList<Soprano::Statement> countryStatements;
        while ( it.next() ) {
            Soprano::Statement s = *it;
            s.setContext( m_countriesGraph );
            countryStatements << s;
            ++cnt;
        }

        m_fetchedCountryStatementsCount += cnt;

        kDebug() << "Fetched" << cnt << "country statements";

        if ( cnt == 0 ) {
            //
            // if we added any statements at all, we finished successfully and can, thus, add
            // the graph metadata
            //
            if ( m_fetchedCountryStatementsCount ) {
                QUrl metaDataGraph = ResourceManager::instance()->generateUniqueUri( "geonames-countries-metadata" );
                m_asyncModel->addStatement( m_countriesGraph, Soprano::Vocabulary::RDF::type(), Soprano::Vocabulary::NRL::InstanceBase(), metaDataGraph );
                m_asyncModel->addStatement( m_countriesGraph, Soprano::Vocabulary::NAO::created(), Soprano::LiteralValue( QDateTime::currentDateTime() ), metaDataGraph );
                m_asyncModel->addStatement( m_countriesGraph, Soprano::Vocabulary::RDFS::label(), Soprano::LiteralValue("geonames/countries"), metaDataGraph );
                m_asyncModel->addStatement( metaDataGraph, Soprano::Vocabulary::RDF::type(), Soprano::Vocabulary::NRL::GraphMetadata(), metaDataGraph );
                m_asyncModel->addStatement( metaDataGraph, Soprano::Vocabulary::NRL::coreGraphMetadataFor(), Soprano::Vocabulary::NRL::GraphMetadata(), metaDataGraph );

                kDebug() << "Successfully fetched all countries.";
                emitResult();
            }
            else {
                kDebug() << "Did not get any country statements";
                setErrorText( i18n( "Failed to get country information from geonames." ) );
                emitResult();
            }
        }
        else {
            m_retryCount = 0;
            m_fetchedCountriesCount += s_countryFetchBatchSize;

            // add the statements asyncroneously
            // FIXME: check result of adding statements
            connect( m_asyncModel->addStatementsAsync( countryStatements ),
                     SIGNAL( resultReady( Soprano::Util::AsyncResult* ) ),
                     this, SLOT( fetchNextBatch() ) );
        }
    }
    else if ( m_retryCount < 5 ) {
        ++m_retryCount;
        fetchNextBatch();
    }
    else {
        kDebug() << "Retried 5 times. Stopping.";
        setErrorText( i18n( "Failed to get country information from geonames." ) );
        emitResult();
    }
}

#include "geonamescountryfetchjob.moc"
