/*
   Copyright (c) 2008 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "pimoannotationplugin.h"
#include "simpleannotation.h"
#include "pimo.h"
#include "pimoplugintools.h"

#include <Soprano/Node>
#include <Soprano/LiteralValue>
#include <Soprano/QueryResultIterator>
#include <Soprano/Util/AsyncModel>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/XMLSchema>

#include <Nepomuk/ResourceManager>
#include <Nepomuk/Thing>
#include <Nepomuk/Variant>
#include <Nepomuk/Types/Literal>
#include <Nepomuk/Types/Property>
#include <Nepomuk/Types/Class>

#include <KPluginFactory>
#include <KPluginLoader>
#include <KDebug>

#include <math.h>

NEPOMUK_EXPORT_ANNOTATION_PLUGIN( Nepomuk::PimoAnnotationPlugin, "nepomuk_pimoannotationplugin" )

namespace {
    /**
     * QVariant::canConvert does not actually try to convert strings to numbers. It for example
     * returns true for strings like "wurst".
     */
    bool canConvert( const QVariant& value, QVariant::Type type )
    {
        if ( value.type() == type ) {
            return true;
        }

        else if ( value.canConvert( type ) ) {
            bool ok = false;
            switch ( type ) {
            case QVariant::Int:
            case QVariant::UInt:
            case QVariant::LongLong:
            case QVariant::ULongLong:
                value.toString().toLongLong( &ok );
                return ok;

            case QVariant::Double:
                value.toString().toDouble( &ok );
                return ok;

            default:
                return true;
            }
        }

        return false;
    }
}



Nepomuk::PimoAnnotationPlugin::PimoAnnotationPlugin( QObject* parent, const QVariantList& )
    : AnnotationPlugin( parent ),
      m_lastPendingResult( 0 )
{
    m_model = new Soprano::Util::AsyncModel( ResourceManager::instance()->mainModel() );
//    m_model->setMode( Soprano::Util::AsyncModel::MultiThreaded ); // looks like there is a bug in the multithread mode - it crashes and is slower
}


Nepomuk::PimoAnnotationPlugin::~PimoAnnotationPlugin()
{
    delete m_model;
}


void Nepomuk::PimoAnnotationPlugin::doGetPossibleAnnotations( const Resource& resource, const QString& filter )
{
    m_resource = resource;

    if ( m_lastPendingResult ) {
        m_lastPendingResult->disconnect( this );
    }

    if ( !filter.isEmpty() ) {
        if ( filter.contains( ':' ) ) {
            QStringList s = filter.split( ':', QString::SkipEmptyParts);
            if ( s.count() == 2 ) {
                m_propertyLabel = s[0];
                m_valueLabel = s[1];
                matchProperty( m_propertyLabel );
                return;
            }
        }
    }

    // done
    emitFinished();
}


void Nepomuk::PimoAnnotationPlugin::matchProperty( const QString& label )
{
    // Step 1: see if we have a direct match to a predicate label
    //         there is no need in selecting unused properties
    QString query = QString( "select distinct ?p where { "
                             "?p <%1> <%2> . "
                             "?p <%3> \"%4\"^^<%5> . "
                             "?x ?p ?y . }" )
                    .arg( Soprano::Vocabulary::RDF::type().toString() )
                    .arg( Soprano::Vocabulary::RDF::Property().toString() )
                    .arg( Soprano::Vocabulary::RDFS::label().toString() )
                    .arg( label )
                    .arg( Soprano::Vocabulary::XMLSchema::string().toString() );
    kDebug() << "Direct match query:" << query;

    m_currentOperation = DirectPropertyMatch;

    m_lastPendingResult = m_model->executeQueryAsync( query, Soprano::Query::QueryLanguageSparql );
    connect( m_lastPendingResult, SIGNAL( resultReady(Soprano::Util::AsyncResult*) ),
             this, SLOT( slotPropertySearchResults( Soprano::Util::AsyncResult*) ) );
}


void Nepomuk::PimoAnnotationPlugin::slotPropertySearchResults( Soprano::Util::AsyncResult* result )
{
    m_lastPendingResult = 0;

    Soprano::QueryResultIterator propertyIt = result->queryResultIterator();
    QList<Nepomuk::Types::Property> properties;
    while ( propertyIt.next() ) {
        properties << propertyIt.binding( "p" ).uri();
        kDebug() << "Found direct match" << propertyIt.binding( "p" ).uri();
    }

    // filter out non-matching properties
    QList<Types::Property>::iterator it = properties.begin();
    while ( it != properties.end() ) {
        if ( !m_resource.pimoThing().hasType( it->domain().uri() ) ) {
            it = properties.erase( it );
        }
        else {
            ++it;
        }
    }

    // if we did not find something yet, go on with the next try
    if ( properties.isEmpty() ) {
        QString query;
        switch( m_currentOperation ) {
        case DirectPropertyMatch: {
            query = QString( "select ?p where { "
                             "?p <%1> <%2> . "
                             "?p <%3> ?label . "
                             "FILTER(REGEX(STR(?label),'%4','i')) . }" )
                    .arg( Soprano::Vocabulary::RDF::type().toString() )
                    .arg( Soprano::Vocabulary::RDF::Property().toString() )
                    .arg( Soprano::Vocabulary::RDFS::label().toString() )
                    .arg( m_propertyLabel );
            kDebug() << "Indirect hit query:" << query;
            m_currentOperation = IndirectPropertyMatchByLabel;
            break;
        }

        case IndirectPropertyMatchByLabel: {
            query = QString( "select ?p where { "
                             "?p <%1> <%2> . "
                             "FILTER(REGEX(STR(?p),'%3','i')) . }" )
                    .arg( Soprano::Vocabulary::RDF::type().toString() )
                    .arg( Soprano::Vocabulary::RDF::Property().toString() )
                    .arg( m_propertyLabel );
            m_currentOperation = IndirectPropertyMatchByName;
            break;
        }

        case IndirectPropertyMatchByName: {
            createRelationAnnotations( properties );
            return;
        }
        }

        m_lastPendingResult = m_model->executeQueryAsync( query, Soprano::Query::QueryLanguageSparql );
        connect( m_lastPendingResult, SIGNAL( resultReady(Soprano::Util::AsyncResult*) ),
                 this, SLOT( slotPropertySearchResults( Soprano::Util::AsyncResult*) ) );
    }
    else {
        createRelationAnnotations( properties );
    }
}


void Nepomuk::PimoAnnotationPlugin::createRelationAnnotations( const QList<Nepomuk::Types::Property>& properties )
{
    m_lastPendingResult = 0;
    m_matchedProperties = properties;

    // IDEA: use subproperties if the domain and range match.
    // FIXME: if the direct match properties did not produce any results below, redo the indirect query

    QVariant value( m_valueLabel );

    if ( !properties.isEmpty() ) {
        //
        // First do the literal relations
        //
        for ( int i = 0; i < properties.count(); ++i ) {
            Types::Property property = properties[i];
            Types::Literal literalRange = property.literalRangeType();
            if ( literalRange.isValid() ) {
                kDebug() << property.uri() << "has literal range of type" << literalRange.dataType();
                if ( canConvert( value, literalRange.dataType() ) ) {
                    SimpleAnnotation* anno = new SimpleAnnotation( m_resource, this );
                    anno->setLabel( "FIXME" );
                    anno->setComment( i18nc( "@action A property (%1) of a selected resource is set to value %2",
                                             "Set '%1' to %2", property.label(), value.toString() ) );
                    anno->setRelevance( 1.0 - calculateDistance( property.label(), m_propertyLabel ) );
                    anno->setProperty( property );
                    anno->setValue( Nepomuk::Variant( value ) );
                    addNewAnnotation( anno );
                }
                else if ( literalRange.dataType() == QVariant::DateTime ) {
                    QString today = i18nc( "@label A fancy way of writing the current date.", "today" );
                    QString yesterday = i18nc( "@label A fancy way of writing a date.", "yesterday" );
                    QString now = i18nc( "@label A fancy way of writing the current time", "now" );
                    if ( !m_valueLabel.compare( today, Qt::CaseInsensitive ) ) {
                        // FIXME: do something
                    }
                    else if ( !m_valueLabel.compare( yesterday, Qt::CaseInsensitive ) ) {
                        // FIXME: do something

                    }
                    else if ( !m_valueLabel.compare( now, Qt::CaseInsensitive ) ) {
                        // FIXME: do something

                    }
                }
            }
        }
    }

    //
    // Find some things to relate to
    //
    QUrl range = Nepomuk::Vocabulary::PIMO::Thing();
    // find things to relate (without inference as always this is a bit hacky)

    // Hint: This query is not exactly what we want as it ignores things without grounding occurrence
    //       but it is way faster.
    // FIXME: 2 queries are faster
    QString query = QString( "select distinct ?r where { "
                             "{ ?r a <%1> . }"
                             " UNION "
                             "{?r a ?t1 . ?t1 <%2> <%1> . }"
                             " UNION "
                             "{?r a ?t2 . ?t2 <%2> ?t3 . ?t3 <%2> <%1> . } . "
                             "?r <%3> ?go . "
                             "{ ?go ?foo ?value . }"
                             " UNION "
                             "{ ?r ?bar ?value . } "
                             "FILTER(ISLITERAL(?value) && REGEX(STR(?value),'%4','i')) . }" )
                    .arg( range.toString() )
                    .arg( Soprano::Vocabulary::RDFS::subClassOf().toString() )
                    .arg( Nepomuk::Vocabulary::PIMO::groundingOccurrence().toString() )
                    .arg( m_valueLabel );
    kDebug() << query;
    m_lastPendingResult = m_model->executeQueryAsync( query, Soprano::Query::QueryLanguageSparql );
    connect( m_lastPendingResult, SIGNAL( resultReady(Soprano::Util::AsyncResult*) ),
             this, SLOT( slotThingSearchFinished( Soprano::Util::AsyncResult*) ) );
}


void Nepomuk::PimoAnnotationPlugin::slotThingSearchFinished( Soprano::Util::AsyncResult* result )
{
    m_lastPendingResult = 0;

    QVariant value( m_valueLabel );

    Soprano::QueryResultIterator it = result->queryResultIterator();
    if ( m_matchedProperties.isEmpty() ) {

        qreal maxResourceAnnotationRelevance = 0.0;

        // 1. Propose to create a new property with a thing that matches m_valueLabel
        //    score higher than the literal ones

        while ( it.next() ) {
            Resource res( it.binding( 0 ).uri() );
            SimpleAnnotation* anno = new SimpleAnnotation( m_resource, this );
            anno->setLabel( "FIXME" );
            anno->setComment( i18nc( "@action A property (%1) of a selected resource is set to value %2",
                                     "Create new relation: Set '%1' to %2", m_propertyLabel, res.genericLabel() ) );
            anno->setRelevance( 0.99 * ( 1.0 - calculateDistance( res.genericLabel(), m_valueLabel ) ) );
            anno->setNewPropertyLabel( m_propertyLabel );
            anno->setValue( Nepomuk::Variant( res ) );
            addNewAnnotation( anno );
            maxResourceAnnotationRelevance = qMax( maxResourceAnnotationRelevance, anno->relevance() );
        }

        // 2. propose the creation of a new property with a literal range (handle int, qreal, and string)

        SimpleAnnotation* anno = new SimpleAnnotation( m_resource, this );
        anno->setLabel( "FIXME" );
        // relevance of creating new stuff is never as good as something matching perfectly
        // and creating a new property linking to resources is always better than linking to literals
        anno->setRelevance( maxResourceAnnotationRelevance > 0.0 ? 0.99*maxResourceAnnotationRelevance : 0.99 );
        anno->setNewPropertyLabel( m_propertyLabel );

        if ( canConvert( value, QVariant::Int ) ) {
            anno->setComment( i18nc("@action A property (%1) of a selected resource is set to value %2",
                                    "Create new relation: Set '%1' to %2", m_propertyLabel, value.toInt() ) );
            anno->setValue( Nepomuk::Variant( value.toInt() ) );
        }
        else if ( canConvert( value, QVariant::Double ) ) {
            anno->setComment( i18nc("@action A property (%1) of a selected resource is set to value %2",
                                    "Create new relation: Set '%1' to %2", m_propertyLabel, value.toDouble() ) );
            anno->setValue( Nepomuk::Variant( value.toDouble() ) );
        }
        else {
            anno->setComment( i18nc("@action A property (%1) of a selected resource is set to value %2",
                                    "Create new relation: Set '%1' to %2", m_propertyLabel, value.toString() ) );
            anno->setValue( Nepomuk::Variant( value.toString() ) );
        }

        // FIXME: if we have no thing to relate to, propose to create a new thing instead of a string value

        addNewAnnotation( anno );

    }

    else { // !m_matchedProperties.isEmpty()
        while ( it.next() ) {
            Resource res( it.binding( 0 ).uri() );
            for ( int i = 0; i < m_matchedProperties.count(); ++i ) {
                Types::Property& property = m_matchedProperties[i];
                if ( res.hasType( property.range().uri() ) ) {
                    SimpleAnnotation* anno = new SimpleAnnotation( m_resource, this );
                    anno->setLabel( "FIXME" );
                    anno->setComment( i18nc( "@action A property (%1) of a selected resource is set to value %2",
                                             "Set '%1' to %2", property.label(), res.genericLabel() ) );
                    anno->setRelevance( ( 1.0 - calculateDistance( property.label(), m_propertyLabel ) )
                                        * ( 1.0 - calculateDistance( res.genericLabel(), m_valueLabel ) ) );
                    anno->setProperty( property );
                    anno->setValue( Nepomuk::Variant( res ) );
                    addNewAnnotation( anno );
                }
                else {
                    kDebug() << "resource" << res.genericLabel() << "does not have type" << property.range().uri() << "for property" << property.uri();
                }
            }
        }
    }

    emitFinished();
}

#include "pimoannotationplugin.moc"
