/*
 * This file is part of the Nepomuk KDE project.
 * Copyright (C) 2008 Adrien Bustany <madcat@mymadcat.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "tags.h"
#include "annotation.h"
#include "pimoplugintools.h"

#include <kiconloader.h>
#include <nepomuk/resource.h>
#include <nepomuk/resourcemanager.h>
#include <nepomuk/tag.h>
#include <soprano/soprano.h>
#include <klocale.h>
#include <KIcon>
#include <KDebug>

#include "annotationresult.h"

#define ICON_NAME "bookmark-toolbar"


namespace Nepomuk {

    class TagAnnotation : public Annotation
    {
    public:
        TagAnnotation( const Resource& resource, const Tag& tag, QObject* parent )
            : Annotation( resource, parent ),
              m_tag( tag ) {
        }

    protected:
        void doCreate() {
            kDebug() << "tagging" << resource().resourceUri() << "with" << m_tag.resourceUri();
            if ( m_tag.label().isEmpty() &&
                !m_tag.identifiers().isEmpty() ) {
                m_tag.setLabel( m_tag.identifiers().first() );
            }
            resource().addTag( m_tag );
            emitFinished();
        }

    private:
        Tag m_tag;
    };

TagsAnnotationPlugin::TagsAnnotationPlugin(QObject* parent, const QList<QVariant>&) : AnnotationPlugin(parent), m_ready(false)
{
	m_soprano = new Soprano::Util::AsyncModel(Nepomuk::ResourceManager::instance()->mainModel());

	m_hasTag = AnnotationProperty(this, Types::Property(Soprano::Vocabulary::NAO::hasTag()), 10);
	m_hasTag.setLabel(i18n("has tag"));
	m_hasTag.setDescription(i18n("attach the tag to the selected file"));
	fetchTags();
}

bool TagsAnnotationPlugin::isReady() const
{
	return m_ready;
}

QString TagsAnnotationPlugin::name() const
{
	return i18n("Tags");
}

QString TagsAnnotationPlugin::iconName() const
{
	return QString(ICON_NAME);
}

AnnotationResult* TagsAnnotationPlugin::getResources(const QString& filter)
{
	AnnotationResult* result = AnnotationPlugin::createResult(this);

	// This function needs the cached data, enqueue the command if we're not ready yet
	if(!isReady()) {
		QHash<QString, QVariant> args;
		args["filter"] = filter;
		enqueueCommand(result, qMakePair(AnnotationPlugin::command_getResources, args));
		return result;
	}

	QList<AnnotationResource> res;
	foreach(const AnnotationResource& r, m_resources) {
		if(r.label().contains(filter, Qt::CaseInsensitive) || r.description().contains(filter, Qt::CaseInsensitive))
			res.append(r);
	}
	result->setResources(res);
	result->setReady();
	return result;
}

AnnotationResult* TagsAnnotationPlugin::getResources(const QString& filter, const Types::Property& /*property*/, const Soprano::Node& /*object*/) {
	// TODO : not implemented here
	AnnotationResult* result = AnnotationPlugin::createResult(this);
	result->setReady();
	return result;
}

AnnotationResult* TagsAnnotationPlugin::getResource(const QUrl& resourceUri)
{
	AnnotationResult* result = AnnotationPlugin::createResult(this);
	// This function needs the cached data, enqueue the command if we're not ready yet
	if(!isReady()) {
		QHash<QString, QVariant> args;
		args["resourceUri"] = resourceUri;
		enqueueCommand(result, qMakePair(AnnotationPlugin::command_getResource, args));
		return result;
	}

	foreach(const AnnotationResource& r, m_resources) {
		if(r.resource().resourceUri() == resourceUri) {
			result->setResources(QList<Nepomuk::AnnotationResource>() << r);
			result->setReady();
			return result;
		}
	}
	result->setReady();
	return result;
}

AnnotationResult* TagsAnnotationPlugin::getResourcesRelatedTo(const QUrl& resourceUri)
{
	AnnotationResult* result = AnnotationPlugin::createResult(this);

	QString query = QString("select ?uid ?label ?description where {"
							"?uid a <%1> ."
							"?uid <%2> ?label ."
							"optional { ?uid <%3> ?description } ."
							"?property <%4> <%5> ."
							"{{?uid ?property <%6>} UNION {<%6> ?property ?uid}} }")
							.arg(Soprano::Vocabulary::NAO::Tag().toString(),
								Soprano::Vocabulary::NAO::prefLabel().toString(),
								Soprano::Vocabulary::NAO::description().toString(),
								Soprano::Vocabulary::RDFS::subPropertyOf().toString(),
								Soprano::Vocabulary::NAO::annotation().toString(),
								resourceUri.toString());
	qDebug() << query;
	Soprano::Util::AsyncResult* sopranoResult = m_soprano->executeQueryAsync(query, Soprano::Query::QueryLanguageSparql);
	connect(sopranoResult, SIGNAL(resultReady(Soprano::Util::AsyncResult*)), this, SLOT(tag_relatedTo_cb(Soprano::Util::AsyncResult*)));

	m_resultsTable[sopranoResult] = result;

	return result;
}

QList<AnnotationProperty> TagsAnnotationPlugin::getProperties(const QString& /*mimetype*/)
{
	QList<AnnotationProperty> properties;
	properties.append(m_hasTag);
	return properties;
}

QList<AnnotationProperty> TagsAnnotationPlugin::getProperties(const QUrl& /*rdfType*/)
{
	return getProperties(QString());
}

void TagsAnnotationPlugin::addRelation(const Types::Property& property, const QUrl& subject, const Soprano::Node& object)
{
	if(property == m_hasTag.property()) {
		Nepomuk::Resource res(subject);
		Nepomuk::Tag tag(object.uri());
		res.addTag(tag);
	}
	emitFinished();
}

void TagsAnnotationPlugin::addRelation(const AnnotationProperty& property, const QUrl& subject, const Soprano::Node& object)
{
	if(property.property() == m_hasTag.property())
		addRelation(property.property(), subject, object);
}

void TagsAnnotationPlugin::fetchTags()
{
	QList<Nepomuk::Tag> tags = Nepomuk::Tag::allTags();
	foreach(const Nepomuk::Tag& tag, tags) {
		m_resources.append(AnnotationResource(this, tag, tag.genericLabel(), tag.genericDescription(), QString(ICON_NAME)));
	}

	m_ready = true;
	executeQueuedCommands();
	emit ready();
}

void TagsAnnotationPlugin::tag_relatedTo_cb(Soprano::Util::AsyncResult* result)
{
	QList<AnnotationResource> res;
	foreach(const Soprano::BindingSet& set, result->value().value<Soprano::QueryResultIterator>().allBindings()) {
		res.append(AnnotationResource (this, set["uid"].uri(), set["label"].literal().toString(), set["description"].literal().toString(), QString(ICON_NAME)));
	}

	AnnotationResult* r = m_resultsTable[result];
	r->setResources(res);
	r->setReady();
	m_resultsTable.remove(result);
}

void TagsAnnotationPlugin::doGetPossibleAnnotations( const Resource& res, const QString& filter )
{
    bool havePerfectMatch = false;
    foreach( const Tag& tag, Tag::allTags() ) {
        QString label = tag.genericLabel();
        if( filter.isEmpty() ||
            label.contains(filter, Qt::CaseInsensitive) ||
            tag.genericDescription().contains(filter, Qt::CaseInsensitive)) {

            // if the resource is already tagged, we want the perfect match
            // handling, too.
            if ( !res.tags().contains( tag ) ) {
                TagAnnotation* anno = new TagAnnotation( res, tag, this );
                anno->setLabel( label );
                anno->setComment( i18nc( "@action Assign a tag to the selected resource",
                                         "Tag with %1", label ) );
                anno->setIcon( KIcon( tag.genericIcon() ) );
                if ( !filter.isEmpty() ) {
                    anno->setRelevance( 1.0 - calculateDistance( label, filter ) );
                }
                addNewAnnotation( anno );
            }

            if ( !filter.isEmpty() &&
                 !filter.compare( label, Qt::CaseInsensitive ) ) {
                havePerfectMatch = true;
            }
        }
    }
    if ( !filter.isEmpty() &&
         !havePerfectMatch ) {
        TagAnnotation* anno = new TagAnnotation( res, Tag( filter ), this );
        anno->setLabel( filter );
        anno->setComment( i18nc( "@action Assign a tag to the selected resource",
                                 "Tag with %1", filter ) );
        anno->setIcon( KIcon( "document-new" ) );
        anno->setRelevance( 0.99 ); // while the relevance is very high, it is lower than perfect matches from other plugins
        addNewAnnotation( anno );
    }

    emitFinished();
}

}

Q_DECLARE_METATYPE(Soprano::QueryResultIterator)

NEPOMUK_EXPORT_ANNOTATION_PLUGIN( Nepomuk::TagsAnnotationPlugin, "nepomuk_tagsannotationplugin" )

#include "tags.moc"
