/*
   Copyright (c) 2008 Sebastian Trueg <trueg@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include "contact.h"
#include "pimo.h"
#include "nco.h"

#include <QtCore/QSharedData>

#include <nepomuk/resourcemanager.h>
#include <nepomuk/resource.h>
#include <nepomuk/thing.h>
#include <nepomuk/variant.h>

#include <Soprano/Model>
#include <Soprano/QueryResultIterator>
#include <Soprano/Node>
#include <Soprano/Vocabulary/XMLSchema>
#include <Soprano/Vocabulary/NAO>


class Nepomuk::PIM::Contact::Private : public QSharedData
{
public:
    Thing pimoThing;
    QList<Resource> occurences;
};



Nepomuk::PIM::Contact::Contact()
    : d( new Private() )
{
}


Nepomuk::PIM::Contact::Contact( const Contact& other )
{
    d = other.d;
}


Nepomuk::PIM::Contact& Nepomuk::PIM::Contact::operator=( const Contact& other )
{
    d = other.d;
    return *this;
}


Nepomuk::PIM::Contact::~Contact()
{
}


bool Nepomuk::PIM::Contact::isValid() const
{
    return d->pimoThing.isValid();
}


Nepomuk::PIM::Contact Nepomuk::PIM::Contact::queryPersonByName( const QString& name, bool create )
{
    Contact c;
    QString query = QString( "select ?thing where { ?thing a <%1> . ?thing <%2> \"%3\"^^<%4> . }" )
                    .arg( Vocabulary::PIMO::Person().toString() )
                    .arg( Soprano::Vocabulary::NAO::prefLabel().toString() )
                    .arg( name )
                    .arg( Soprano::Vocabulary::XMLSchema::string().toString() );
    Soprano::QueryResultIterator it = ResourceManager::instance()->mainModel()->executeQuery( query, Soprano::Query::QueryLanguageSparql );
    if( it.next() ) {
        c.d->pimoThing = it.binding( "thing" ).uri();
        // FIXME: what about pimo:occurence, pimo:referencingOccurence?
        c.d->occurences = c.d->pimoThing.property( Vocabulary::PIMO::groundingOccurrence() ).toResourceList();
    }
    else if( create ) {
        c.d->pimoThing = Resource( QUrl(), Vocabulary::PIMO::Person() );
        c.d->pimoThing.setLabel( name );
    }
    return c;
}


Nepomuk::PIM::Contact Nepomuk::PIM::Contact::queryPersonByEmail( const QString& emailAddress, bool create )
{
    Contact c;

    QString query = QString( "select ?person ?thing where { "
                             "?person a <%1> . "
                             "?person <%2> ?email . "
                             "?email <%3> \"%4\"^^<%5> . }" )
                    .arg( Vocabulary::NCO::PersonContact().toString() )
                    .arg( Vocabulary::NCO::hasEmailAddress().toString() )
                    .arg( Vocabulary::NCO::emailAddress().toString() )
                    .arg( emailAddress )
                    .arg( Soprano::Vocabulary::XMLSchema::string().toString() );
    Soprano::QueryResultIterator it = ResourceManager::instance()->mainModel()->executeQuery( query, Soprano::Query::QueryLanguageSparql );
    if ( it.next() ) {
        QUrl person = it.binding( "person" ).uri();
        it.close();
        c.d->occurences.append( Resource( person ) );
        c.d->pimoThing = c.d->occurences.first().pimoThing();
    }
    else if ( create ) {
        Resource person( QUrl(), Vocabulary::NCO::PersonContact() );
        Resource email( QUrl( "mailto:/" + emailAddress ), Vocabulary::NCO::EmailAddress() );
        email.setProperty( Vocabulary::NCO::emailAddress(), Variant( emailAddress ) );
        person.setProperty( Vocabulary::NCO::hasEmailAddress(), email );
        c.d->occurences.append( person );
        c.d->pimoThing = c.d->occurences.first().pimoThing();
    }
    c.d->pimoThing.addType( Vocabulary::PIMO::Person() );
    return c;
}


Nepomuk::PIM::Contact Nepomuk::PIM::Contact::queryPersonByIdentifier( const QUrl& idType, const QString& identifier, bool create )
{
    Contact c;

//     QString query = QString( "select ?thing where { ?thing <%1> ?go . ?go <%2> \"%3\"^^<%4> . }" )
//                     .arg( Vocabulary::PIMO::groudingOccurence().toString() )
//                     .arg( QString::fromAscii( idType.toEncoded() ) )
//                     .arg( identifier )
//                     .arg( Soprano::Vocabulary::XMLSchema::string() );
    QString query = QString( "select ?thing where { ?thing <%1> \"%2\"^^<%3> . }" )
                    .arg( QString::fromAscii( idType.toEncoded() ) )
                    .arg( identifier )
                    .arg( Soprano::Vocabulary::XMLSchema::string().toString() );
    Soprano::QueryResultIterator it = ResourceManager::instance()->mainModel()->executeQuery( query, Soprano::Query::QueryLanguageSparql );
    if( it.next() ) {
        c.d->pimoThing = it.binding( "thing" ).uri();
        // FIXME: what about pimo:occurence, pimo:referencingOccurence?
        c.d->occurences = c.d->pimoThing.property( Vocabulary::PIMO::groundingOccurrence() ).toResourceList();
    }
    else if( create ) {
        c.d->pimoThing = Resource( QUrl(), Vocabulary::PIMO::Person() );
        c.d->pimoThing.setProperty( idType, identifier );
    }
    return c;
}


QString Nepomuk::PIM::Contact::name() const
{
    if( !d->pimoThing.label().isEmpty() ) {
        return d->pimoThing.label();
    }
    else {
        QStringList names;
        // FIXME: do only use the first name found
        names += property( QUrl("http://www.semanticdesktop.org/ontologies/2007/03/22/nco#nameGiven") ).toString();
        names += property( QUrl("http://www.semanticdesktop.org/ontologies/2007/03/22/nco#nameAdditional") ).toStringList();
        names += property( QUrl("http://www.semanticdesktop.org/ontologies/2007/03/22/nco#nameFamily" )).toString();
        return names.join( " " );
    }
}


void Nepomuk::PIM::Contact::setName( const QString& name )
{
    d->pimoThing.setLabel( name );
}


QString Nepomuk::PIM::Contact::emailAddress() const
{
    // FIXME: check the grouding occurences
    QList<Resource> el = property( QUrl::fromEncoded( "http://www.semanticdesktop.org/ontologies/2007/03/22/nco#hasEmailAddress",
                                                      QUrl::StrictMode ) ).toResourceList();
    if( !el.isEmpty() ) {
        return el.first().property( QUrl::fromEncoded( "http://www.semanticdesktop.org/ontologies/2007/03/22/nco#emailAddress",
                                                       QUrl::StrictMode ) ).toString();
    }
    else {
        return QString();
    }
}


QString Nepomuk::PIM::Contact::identifier( const QUrl& idTyoe ) const
{
    QStringList ids = property( idTyoe ).toStringList();
    return( ids.isEmpty() ? QString() : ids.first() );
}


Nepomuk::Variant Nepomuk::PIM::Contact::property( const QUrl& property ) const
{
    if( d->pimoThing.hasProperty( property ) ) {
        return d->pimoThing.property( property );
    }
    else {
        Variant v;
        foreach( const Resource& res, d->occurences ) {
            if( res.hasProperty( property ) ) {
                v.append( res.property( property ) );
            }
        }
        return v;
    }
}


Nepomuk::Thing Nepomuk::PIM::Contact::pimoThing() const
{
    return d->pimoThing;
}


QList<Nepomuk::Resource> Nepomuk::PIM::Contact::incarnations() const
{
    return d->occurences;
}


void Nepomuk::PIM::Contact::addIncarnation( const QUrl& idType, const QString& id )
{
    Nepomuk::Resource c( QUrl(), Vocabulary::NCO::PersonContact() );
    c.setProperty( idType, id );
    Nepomuk::Variant v = d->pimoThing.property( Vocabulary::PIMO::groundingOccurrence() );
    v.append( c );
    d->pimoThing.setProperty( Vocabulary::PIMO::groundingOccurrence(), v );
}
