/*
 *
 * $Id: sourceheader 511311 2006-02-19 14:51:05Z trueg $
 *
 * This file is part of the Nepomuk KDE project.
 * Copyright (C) 2008 Sebastian Trueg <trueg@kde.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include "kio_nepomuk.h"

#include <QtCore/QByteArray>
#include <QtCore/QTextStream>
#include <QtCore/QDateTime>

#include <KComponentData>
#include <KDebug>
#include <KUser>
#include <KMessageBox>

#include <nepomuk/resource.h>
#include <nepomuk/resourcemanager.h>
#include <nepomuk/variant.h>
#include <nepomuk/class.h>
#include <nepomuk/property.h>

#include <Soprano/Model>
#include <Soprano/Node>
#include <Soprano/Statement>
#include <Soprano/StatementIterator>
#include <Soprano/Vocabulary/RDF>
#include <Soprano/Vocabulary/NAO>


Nepomuk::NepomukProtocol::NepomukProtocol( const QByteArray& poolSocket, const QByteArray& appSocket )
    : KIO::SlaveBase( "nepomuk", poolSocket, appSocket )
{
    ResourceManager::instance()->init();
}


Nepomuk::NepomukProtocol::~NepomukProtocol()
{
}

namespace {
    QString typesToHtml( const QList<QUrl>& types )
    {
        QList<Nepomuk::Types::Class> typeClasses;
        foreach( const QUrl& type, types ) {
            typeClasses << Nepomuk::Types::Class( type );
        }

        // remove all types that are supertypes of others in the list
        QList<Nepomuk::Types::Class> normalizedTypes;
        for ( int i = 0; i < typeClasses.count(); ++i ) {
            Nepomuk::Types::Class& type = typeClasses[i];
            bool use = true;
            for ( int j = 0; j < typeClasses.count(); ++j ) {
                if ( type != typeClasses[j] &&
                     typeClasses[j].isSubClassOf( type ) ) {
                    use = false;
                    break;
                }
            }
            if ( use ) {
                normalizedTypes << type;
            }
        }

        // extract the labels
        QStringList typeStrings;
        for ( int i = 0; i < normalizedTypes.count(); ++i ) {
            typeStrings << normalizedTypes[i].label();
        }

        return typeStrings.join( ", " );
    }
}


void Nepomuk::NepomukProtocol::get( const KUrl& url )
{
    kDebug() << url;

    QString action;
    if ( url.hasQuery() ) {
        QMap<QString, QString> qi = url.queryItems();
        for( QMap<QString, QString>::const_iterator it = qi.constBegin();
             it != qi.constEnd(); ++it ) {
            if ( it.key() == "action" ) {
                action = it.value();
            }
            else {
                error( KIO::ERR_MALFORMED_URL, url.url() );
                return;
            }
        }
    }

    if ( !action.isEmpty() ) {
        KUrl nonQueryUrl( url );
        nonQueryUrl.setQuery( QString() );
        handleAction( nonQueryUrl, action );
        return;
    }

    kDebug() << "Emitting mimetype before data";
    mimeType( "text/html" );

    Nepomuk::Resource res( url );
    bool exists = res.exists();

    QByteArray output;

    QString label = res.genericLabel();

    QTextStream os( &output, QIODevice::WriteOnly );
    os << "<html xmlns=\"http://www.w3.org/1999/xhtml\" version=\"-//W3C//DTD XHTML 1.2//EN\" xml:lang=\"en\">"
       << "<head>"
       << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />"
       << "<title>" << label << "</title>"
       << "<style>"
        "body { "
        "  margin: 0; "
        "  padding: 0; "
        "  text-align: center; "
        "  font-size: 0.8em; "
        "  font-family: \"Bitstream Vera Sans\", \"Lucida Grande\", \"Trebuchet MS\", sans-serif; "
        "  color: #535353; "
        "  background: #ffffff; } "
        "table { "
        "  font-size: 1.0em; "
        "} "
        "h1 { "
        "  padding: 0; "
        "  text-align: left; "
        "  font-weight: bold; "
        "  color: #f7800a; "
        "  background: transparent; } "
        "h1 { "
        "  margin: 0 0 0.3em 0; "
        "  font-size: 1.7em; } "
        "h2, h3, h4 { "
        "  margin: 1.3em 0 0 0.3em } "
        "h2 { "
        "  font-size: 1.2em; } "
        "h3 { font-size: 1.4em; } "
        "h4 { font-size: 1.3em; } "
        "h5 { font-size: 1.2em; } "
        "a:link { "
        "  padding-bottom: 0; "
        "  text-decoration: none; "
        "  color: #0057ae; } "
        "a:visited { "
        "  padding-bottom: 0; "
        "  text-decoration: none; "
        "  color: #644A9B; } "
        "a[href]:hover { "
        "  text-decoration: underline; }"
        "p { "
        "  margin-top: 0.5em; "
        "  margin-bottom: 0.9em; "
        "  text-align: justify; } "
        "li { "
        "  text-align: left; } "
        ". clearer { "
        "  clear: both; "
        "  height: 1px; } "
        "#content { "
        "  width: 100%; } "
        "#main { "
        "  padding: 10px; "
        "  text-align: left; } "
        "#body_wrapper { "
        "  margin: 0 auto; "
        "  width: 60em; "
        "  min-width: 770px; "
        "  max-width: 45em; "
        "  border: #bcbcbc solid; "
        "  border-width: 0 0 0 1px; } "
        "#body { "
        "  float: left; "
        "  margin: 0; "
        "  padding: 0; "
        "  min-height: 40em; "
        "  width: 60em; "
        "  min-width: 770px; "
        "  max-width: 45em; } "
        "#relations { "
        "  padding: 0 20 0 20; }"
        "</style></head>"
       << "<body><div id=\"body_wrapper\"><div id=\"body\"><div class=\"content\"><div id=\"main\"><div class=\"clearer\">&nbsp;</div>";

    os << "<h1>" << label << "</h1>"
       << "Type: " << ( exists ? typesToHtml( res.types() ) : i18n( "Resource does not exist" ) );

    os << "<h2>Relations:</h2><div id=\"relations\"><table>";

    // query relations
    Soprano::StatementIterator it = ResourceManager::instance()->mainModel()->listStatements( url, Soprano::Node(), Soprano::Node() );
    while ( it.next() ) {
        Soprano::Statement s = it.current();
        if ( s.predicate().uri() != Soprano::Vocabulary::RDF::type() ) {
            Nepomuk::Types::Property p( s.predicate().uri() );
            os << "<tr><td align=right><i>" << ( p.label().isEmpty() ? p.name() : p.label() ) << "</i></td><td width=16px></td><td>";
            if ( s.object().isLiteral() ) {
                if ( s.object().literal().isDateTime() )
                    os << KGlobal::locale()->formatDateTime( s.object().literal().toDateTime(), KLocale::FancyShortDate );
                else
                    os << s.object().toString();
            }
            else {
                Resource resource( s.object().uri() );
                os << QString( "<a href=\"%1\">%2</a> (%3)" )
                    .arg( QString::fromAscii( resource.resourceUri().toEncoded() ) )
                    .arg( resource.genericLabel() )
                    .arg( typesToHtml( resource.types() ) );
            }
            os << "</td></tr>";
        }
    }
    os << "</table></div>";

    // query backlinks
    os << "<h2>Backlinks:</h2><div id=\"relations\"><table>";
    Soprano::StatementIterator itb = ResourceManager::instance()->mainModel()->listStatements( Soprano::Node(), Soprano::Node(), url );
    while ( itb.next() ) {
        Soprano::Statement s = itb.current();
        if ( s.predicate().uri() != Soprano::Vocabulary::RDF::type() ) {
            Resource resource( s.subject().uri() );
            Nepomuk::Types::Property p( s.predicate().uri() );
            os << "<td align=right>"
               << QString( "<a href=\"%1\">%2</a> (%3)" )
                .arg( QString::fromAscii( s.subject().uri().toEncoded() ) )
                .arg( resource.genericLabel() )
                .arg( typesToHtml( resource.types() ) )
               << "</td>"
               << "<td width=16px></td>"
               << "<td><i>" << ( p.label().isEmpty() ? p.name() : p.label() ) << "</i></td></tr>";
        }
    }
    os << "</table></div>";

    if ( exists ) {
        os << "<h2>Actions:</h2>"
           << "<div id=\"relations\"><a href=\"" << url.url() << "?action=delete\">" << i18n( "Delete resource" ) << "</a></div>";
    }

    os << "</div></div></div></div></body></html>";
    os.flush();

    kDebug() << "created data:" << output;

    data( output );
    finished();
}


void Nepomuk::NepomukProtocol::stat( const KUrl& url )
{
    kDebug() << url;

    KIO::UDSEntry uds;

    Nepomuk::Resource res( url );

    QString label = res.genericLabel();
    QString name = label;

    // make sure name is not the URI (which is the fallback of genericLabel() and will lead to crashes in KDirModel)
    if ( name.contains( '/' ) ) {
        name = name.section( '/', -1 );
        if ( name.isEmpty() )
            name = res.resourceUri().fragment();
        if ( name.isEmpty() )
            name = res.resourceUri().toString().replace( '/', '_' );
    }

    uds.insert( KIO::UDSEntry::UDS_NAME, name );
    uds.insert( KIO::UDSEntry::UDS_DISPLAY_NAME, label );
    uds.insert( KIO::UDSEntry::UDS_CREATION_TIME, res.property( Soprano::Vocabulary::NAO::created() ).toDateTime().toTime_t() );
    uds.insert( KIO::UDSEntry::UDS_MODIFICATION_TIME, res.property( Soprano::Vocabulary::NAO::lastModified() ).toDateTime().toTime_t() );
    uds.insert( KIO::UDSEntry::UDS_ACCESS, 0700 );
    uds.insert( KIO::UDSEntry::UDS_USER, KUser().loginName() );
    uds.insert( KIO::UDSEntry::UDS_MIME_TYPE, "application/x-nepomuk-resource" );
    QString icon = res.genericIcon();
    if ( !icon.isEmpty() ) {
        uds.insert( KIO::UDSEntry::UDS_ICON_NAME, icon );
    }

    statEntry( uds );

    finished();
}


void Nepomuk::NepomukProtocol::mimetype( const KUrl& url )
{
    kDebug() << url;

    // for now we do the simple thing
    // FIXME: later, forward tags to the tags kioslave and so on
    mimeType( "text/html" );
    finished();
}


void Nepomuk::NepomukProtocol::handleAction( const KUrl& url, const QString& action )
{
    if ( action == "delete" ) {
        Nepomuk::Resource res( url );
        if ( res.exists() ) {
            if ( messageBox( i18n( "Do you really want to delete the resource and all relations "
                                   "to and from it?" ),
                             KIO::SlaveBase::QuestionYesNo,
                             i18n( "Delete Resource" ) ) == KMessageBox::Yes ) {
                res.remove();
            }
            kDebug() << "Redirect to" << url << "after deletion";
            redirection( url );
            finished();
        }
        else {
            error( KIO::ERR_DOES_NOT_EXIST, url.url() );
        }
    }
    else {
        error( KIO::ERR_MALFORMED_URL, url.url() );
    }
}

extern "C"
{
    int KDE_EXPORT kdemain( int argc, char **argv ) {

        KComponentData componentData("kio_nepomuk");

        if (argc != 4)
        {
            fprintf(stderr, "Usage: kio_nepomuk protocol domain-socket1 domain-socket2\n");
            exit(-1);
        }

        Nepomuk::NepomukProtocol slave(argv[2], argv[3]);
        slave.dispatchLoop();

        return 0;
    }
}
