/*
 * This file is part of the Nepomuk KDE project.
 * Copyright (C) 2008 Adrien Bustany <madcat@mymadcat.com>
 *               2008 Sebastian Trueg <trueg@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "nepomukmenuplugin.h"
#include "newtagdialog.h"

#include <konq_popupmenuinformation.h>
#include <KHTMLView>
#include <KHTMLPart>
#include <dom/html_document.h>
#include <dom/dom_string.h>
#include <KTemporaryFile>
#include <KDebug>
#include <KLocalizedString>
#include <KActionMenu>

#include <QtGui/QWidgetAction>
#include <QtDBus/QDBusMessage>
#include <QtDBus/QDBusConnection>

#include <Nepomuk/ResourceManager>
#include <Nepomuk/Tag>
#include <Nepomuk/Thing>
#include <Nepomuk/Variant>
#include <Nepomuk/KRatingWidget>


#define PIMO_WEBPAGE_TYPE "http://www.semanticdesktop.org/ontologies/2007/11/01/pimo#Document"
#define PIMO_GROUNDINGOCCURRENCE "http://www.semanticdesktop.org/ontologies/2007/11/01/pimo#groundingOccurrence"
#define NFO_WEBSITE_TYPE "http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#Website"

Q_DECLARE_METATYPE( Nepomuk::Tag )

NepomukMenuPlugin::NepomukMenuPlugin( QObject* parent, const QVariantList& )
    : KonqPopupMenuPlugin( parent )
{
    Nepomuk::ResourceManager::instance()->init();

	m_rateAction = new QWidgetAction( this );
	m_ratingWidget = new KRatingWidget();
	m_rateAction->setDefaultWidget( m_ratingWidget );
	connect( m_ratingWidget, SIGNAL(ratingChanged(unsigned int)),
             this, SLOT(ratingChanged(unsigned int)) );

	m_addTagAction = new KAction( i18nc("@action:inmenu", "Add a new tag..."), this );
	connect( m_addTagAction, SIGNAL(triggered(bool)),
             this, SLOT(addNewTag()) );
}


NepomukMenuPlugin::~NepomukMenuPlugin()
{
}


void NepomukMenuPlugin::setup( KActionCollection* collection, const KonqPopupMenuInformation& info, QMenu* menu )
{
    // We only want 1 item selections
	if( info.items().size() != 1 ) {
        return;
    }

    m_currentUri = info.items().urlList().first();

    Nepomuk::Resource nfoResource( m_currentUri, QUrl::fromEncoded(NFO_WEBSITE_TYPE, QUrl::StrictMode) );
    Nepomuk::Thing webPageThing = nfoResource.pimoThing();

	// All actions in menu will be deleted by the framework

    // Rating action
	KActionMenu* rateMenu = new KActionMenu( i18nc("@action:inmenu", "Rate this page"), collection );
	rateMenu->addAction(m_rateAction);
	menu->addAction(rateMenu);
    m_ratingWidget->setRating(webPageThing.rating());


	// Tagging action
	KActionMenu* tagMenu = new KActionMenu( i18nc("@title:menu", "Tag this page"), collection );
	QList<Nepomuk::Tag> allTags = Nepomuk::Tag::allTags();
    QList<Nepomuk::Tag> webPageTags = webPageThing.tags();
	foreach(const Nepomuk::Tag& tag, allTags) {
		KAction* action = new KAction(tag.genericLabel(), tagMenu); // Will be destroyed when the menu gets closed
		action->setCheckable(true);
        action->setData( QVariant::fromValue( tag ) );
		tagMenu->addAction(action);
        if ( webPageTags.contains( tag ) ) {
            action->setChecked( true );
        }
		connect( action, SIGNAL(toggled(bool)), this, SLOT(slotTagToggled(bool)) );
	}
	tagMenu->addAction(m_addTagAction);
	menu->addAction(tagMenu);

	m_view = qobject_cast<KHTMLView*>(info.parentWidget());
}


Nepomuk::Resource NepomukMenuPlugin::saveCurrentPage()
{
	Nepomuk::Resource nfoResource( m_currentUri, QUrl::fromEncoded(NFO_WEBSITE_TYPE, QUrl::StrictMode) );
    Nepomuk::Thing webPageThing = nfoResource.pimoThing();
    webPageThing.addType( QUrl::fromEncoded(PIMO_WEBPAGE_TYPE, QUrl::StrictMode) );

	// Try to get the page title by accessing the KHtml component
	if( m_view ) {
        if ( webPageThing.label().isEmpty() ) {
            webPageThing.setLabel( m_view->part()->htmlDocument().title().string() );
        }

        // send the page to Strigi for full text search on annotated webpages
		KTemporaryFile tmpFile;
		tmpFile.setAutoRemove( false );
		tmpFile.open();
		QTextStream s( &tmpFile );
		s << m_view->part()->htmlDocument().toHTML();
		QDBusMessage msg = QDBusMessage::createMethodCall( "org.kde.nepomuk.services.nepomukstrigiservice",
														   "/nepomukstrigiservice",
														   "org.kde.nepomuk.Strigi",
														   "analyzeResourceFromTempFileAndDeleteTempFile" );
		msg.setArguments( QList<QVariant>()
						  << QString::fromAscii( nfoResource.resourceUri().toEncoded() )
						  << QDateTime::currentDateTime().toTime_t()
						  << tmpFile.fileName() );
		QDBusConnection::sessionBus().send( msg );
	}

	return webPageThing;
}


void NepomukMenuPlugin::ratingChanged(unsigned int rating)
{
	saveCurrentPage().setRating(rating);
}


void NepomukMenuPlugin::slotTagToggled( bool checked )
{
	Nepomuk::Resource res = saveCurrentPage();
    Nepomuk::Tag toggledTag = static_cast<QAction*>( sender() )->data().value<Nepomuk::Tag>();
    if ( checked ) {
        res.addTag( toggledTag );
    }
    else {
        QList<Nepomuk::Tag> tags = res.tags();
        tags.removeAll( toggledTag );
        res.setTags( tags );
    }
}


void NepomukMenuPlugin::addNewTag()
{
    Nepomuk::Tag newTag = NewTagDialog::createTag();
    if ( newTag.isValid() ) {
        saveCurrentPage().addTag( newTag );
    }
}

#include "nepomukmenuplugin.moc"
